

//////////////////////////////////////////////////////////////////
//                                                              //
//           PLINK (c) 2005-2006 Shaun Purcell                  //
//                                                              //
// This file is distributed under the GNU General Public        //
// License, Version 2.  Please see the file COPYING for more    //
// details                                                      //
//                                                              //
//////////////////////////////////////////////////////////////////


#include <iostream>
#include <iomanip>
#include <climits>
#include <fstream>
#include <bitset>
#include <map>
#include "plink.h"
#include "options.h"
#include "perm.h"
#include "sets.h"
#include "helper.h"

extern ofstream LOG;

void Plink::printLOG(string s)
{
  LOG << s;
  LOG.flush();
  
  if (!par::silent)
    {
      cout << s;
      cout.flush();
    }
}

void Plink::displayGenomePV()
{

  ofstream PLO;
  string f = par::output_file_name + ".plink2";
  PLO.open(f.c_str(),ios::out);
  
  printLOG("Writing genome-wide corrected PLINK results to [ " +f +" ] \n");
  
  for (int i=0; i<original.size();i++)
    {
      for (int j=0; j<original[i].size(); j++)
	{
	  double pv=0;
	  for (int k=0; k<maxr2.size(); k++)
	    if (maxr2[k] >= original[i][j]) pv++;
	  PLO << "G " 
	      << double(pv+1)/double(par::replicates+1)
	      << "\n";
	}
    }
  PLO.close();
}


void Plink::display_pairList()
{

  if (par::SNP_major)
    SNP2Ind();
  
  string f = par::output_file_name + ".plist";
  
  printLOG("Writing pair-list file to [ " + f + " ] \n");
  ofstream PED(f.c_str(), ios::out);
  PED.clear();
  PED.precision(4);

  // Find individuals
  Individual * p1 = NULL;
  Individual * p2 = NULL;
  int i1, i2;

  for (int i=0; i<n; i++)
    {
      if ( sample[i]->fid == par::plist_fid1 && sample[i]->iid == par::plist_iid1 )
	{ p1 = sample[i]; i1 = i; }
      if ( sample[i]->fid == par::plist_fid2 && sample[i]->iid == par::plist_iid2 )
	{ p2 = sample[i]; i2 = i; } 
    }
  
  if (p1 == NULL || p2 == NULL)
    error("Problem finding individuals indicated in --plist\n");


  PED << setw(4) << "CHR" << " " 
      << setw(par::pp_maxsnp) << "SNP" << " "
      << setw(14) << "BP" << " "
      << setw(par::pp_maxfid+par::pp_maxiid) << (p1->fid+"/"+p1->iid) << " "
      << setw(par::pp_maxfid+par::pp_maxiid) << (p2->fid+"/"+p2->iid) << " "
      << setw(4) << "A1" << " "
      << setw(8) << "MAF" << " "
      << setw(4) << "IBS" << "\n";
  
  for (int l=0; l<nl_all; l++)  
    {
      PED << setw(4) << locus[l]->chr << " "
	  << setw(par::pp_maxsnp) << locus[l]->name << " "
	  << setw(14) << locus[l]->bp << " "
	  << setw(par::pp_maxfid+par::pp_maxiid) << genotype(*this,i1,l) << " "
	  << setw(par::pp_maxfid+par::pp_maxiid) << genotype(*this,i2,l) << " "
	  << setw(4) << locus[l]->allele1 << " "
	  << setw(8) << locus[l]->freq << " ";
      
      bool a1 = p1->one[l];
      bool a2 = p1->two[l];

      bool b1 = p2->one[l];
      bool b2 = p2->two[l];
	
      bool ibs0 = false;
      bool ibs1 = false;
      bool miss = false;
      
      if ( a1 == a2 && 
	   b1 == b2 && 
	   a1 != b1  ) ibs0 = true;
      else if ( a1 && !(a2) ) miss = true;
      else if ( b1 && !(b2) ) miss = true;
      else if ( a1 != b1 ||
		a2 != b2 ) ibs1 = true;

      PED << setw(4);
      if (ibs0) PED << "0" << "\n";
      else if (ibs1) PED << "1" << "\n";
      else if (miss) PED << "NA" << "\n";
      else PED << "2" << "\n";
      
    }
  
  PED.close();

}

void Plink::display_listByAllele()
{

  if (!par::SNP_major)
    Ind2SNP();

  // Create an output file that lists one allele per line
  // N = number of individuals FID1 IID2 FID2 IID2 ... FIDN IIDN 
  
  // SNP ALLELE N { Individual list ( 2N entries) } 

  string f = par::output_file_name + ".recode.list";
  
  printLOG("Writing recoded list file to [ " + f + " ] \n");
  ofstream PED(f.c_str(), ios::out);
  PED.clear();

  vector<CSNP*>::iterator s = SNP.begin();
  vector<Locus*>::iterator loc = locus.begin();

  while ( s != SNP.end() )
    {
      // Genotype 11
      PED << (*loc)->chr << par::recode_delimit
	  << (*loc)->name << par::recode_delimit
	  << (*loc)->allele1 << (*loc)->allele1 ;
      
      vector<bool>::iterator i1 = (*s)->one.begin();
      vector<bool>::iterator i2 = (*s)->two.begin();
      vector<Individual*>::iterator gperson = sample.begin();
      
      while ( gperson != sample.end() )
	{
	  if ( (!*i1) && (!*i2) ) 
	    PED << par::recode_delimit << (*gperson)->fid 
		<< par::recode_delimit << (*gperson)->iid ;
	  i1++;
	  i2++;
	  gperson++;
	}
      PED << "\n";
      
      // Genotype 12
      PED << (*loc)->chr << par::recode_delimit
	  << (*loc)->name << par::recode_delimit
	  << (*loc)->allele1 << (*loc)->allele2 ;
      i1 = (*s)->one.begin();
      i2 = (*s)->two.begin();
      gperson = sample.begin();
      while ( gperson != sample.end() )
	{
	  if ( (!*i1) && *i2 ) 
	    PED << par::recode_delimit << (*gperson)->fid 
		<< par::recode_delimit << (*gperson)->iid ;
	  i1++;
	  i2++;
	  gperson++;
	}
      PED << "\n";

      // Genotype 22
      PED << (*loc)->chr << par::recode_delimit
	  << (*loc)->name << par::recode_delimit
	  << (*loc)->allele2 << (*loc)->allele2 ;
      i1 = (*s)->one.begin();
      i2 = (*s)->two.begin();
      gperson = sample.begin();
      while ( gperson != sample.end() )
	{
	  if ( *i1 && *i2 ) 
	    PED << par::recode_delimit << (*gperson)->fid 
		<< par::recode_delimit << (*gperson)->iid ;
	  i1++;
	  i2++;
	  gperson++;
	}
      PED << "\n";

      // Genotype 00
      PED << (*loc)->chr << par::recode_delimit
	  << (*loc)->name << par::recode_delimit
	  << "00";
      i1 = (*s)->one.begin();
      i2 = (*s)->two.begin();
      gperson = sample.begin();
      while ( gperson != sample.end() )
	{
	  if ( *i1 && !*i2 ) 
	    PED << par::recode_delimit << (*gperson)->fid 
		<< par::recode_delimit << (*gperson)->iid ;
	  i1++;
	  i2++;
	  gperson++;
	}
      PED << "\n";
      
      
      // Next SNP
      s++;
      loc++;
    }
  
  PED.close();

}

void make2LTable(ofstream & TWOL, Plink & P, int m1, int m2, vector_t count, bool percent)
{
  vector_t margA(4);
  vector_t margB(4);
  double total = 0;
  
  for (int i=0; i<16; i++)
    total += count[i];

  if (percent)
    {
      for (int i=0; i<16; i++)
	count[i] /= total;
      total = 1;
    }

  margA[0] = count[0]+count[1]+count[2]+count[3];
  margA[1] = count[4]+count[5]+count[6]+count[7];
  margA[2] = count[8]+count[9]+count[10]+count[11];
  margA[3] = count[12]+count[13]+count[14]+count[15];

  margB[0] = count[0]+count[4]+count[8]+count[12];
  margB[1] = count[1]+count[5]+count[9]+count[13];
  margB[2] = count[2]+count[6]+count[10]+count[14];
  margB[3] = count[3]+count[7]+count[11]+count[15];


  TWOL << setw(par::pp_maxsnp) << " " << " "
       << setw(4) << " " << " "
       << "   " << par::twolocus_snp2 << "\n";
  
  TWOL << setw(par::pp_maxsnp) << " " << " "
       << setw(4) << " " << " "
       << setw(6) << (P.locus[m2]->allele1+"/"+P.locus[m2]->allele1) << " "
       << setw(6) << (P.locus[m2]->allele1+"/"+P.locus[m2]->allele2) << " "
       << setw(6) << (P.locus[m2]->allele2+"/"+P.locus[m2]->allele2) << " "
       << setw(6) << "0/0" << " "
       << setw(6) << "*/*" << "\n";

  TWOL << setw(par::pp_maxsnp) << par::twolocus_snp1 << " "
       << setw(4) << (P.locus[m1]->allele1+"/"+P.locus[m1]->allele1) << " ";

  TWOL << setw(6) << count[0] << " "
       << setw(6) << count[1] << " "
       << setw(6) << count[2] << " "
       << setw(6) << count[3] << " "
       << setw(6) << margA[0] << "\n";

  TWOL << setw(par::pp_maxsnp) << " " << " "
       << setw(4) << (P.locus[m1]->allele1+"/"+P.locus[m1]->allele2) << " ";

  TWOL << setw(6) << count[4] << " "
       << setw(6) << count[5] << " "
       << setw(6) << count[6] << " "
       << setw(6) << count[7] << " "
       << setw(6) << margA[1] << "\n";

  TWOL << setw(par::pp_maxsnp) << " " << " "
       << setw(4) << (P.locus[m1]->allele2+"/"+P.locus[m1]->allele2) << " ";

  TWOL << setw(6) << count[8] << " "
       << setw(6) << count[9] << " "
       << setw(6) << count[10] << " "
       << setw(6) << count[11] << " "
       << setw(6) << margA[2] << "\n";

  TWOL << setw(par::pp_maxsnp) << " " << " "
       << setw(4) << "0/0" << " ";

  TWOL << setw(6) << count[12] << " "
       << setw(6) << count[13] << " "
       << setw(6) << count[14] << " "
       << setw(6) << count[15] << " "
       << setw(6) << margA[3] << "\n";

  TWOL << setw(par::pp_maxsnp) << " " << " "
       << setw(4) << "*/*" << " ";

  TWOL << setw(6) << margB[0] << " "
       << setw(6) << margB[1] << " "
       << setw(6) << margB[2] << " "
       << setw(6) << margB[3] << " "
       << setw(6) << total << "\n";

  TWOL << "\n";

}

void Plink::display_twolocus()
{
  
  if (!par::SNP_major)
    Ind2SNP();

  // Find the two SNPs
  
  // Write to file a contingency table, possibly stratified 
  // by phenotype

  string f = par::output_file_name + ".twolocus";
  
  printLOG("Writing two-locus table for " 
	   + par::twolocus_snp1 + " by " 
	   + par::twolocus_snp2 + " to [ " + f + " ] \n");

  ofstream TWOL(f.c_str(), ios::out);
  TWOL.clear();
  TWOL.precision(3);

  int m1 = getMarkerNumber(*this,par::twolocus_snp1);
  int m2 = getMarkerNumber(*this,par::twolocus_snp2);
  
  if (m1<0) error("Marker "+par::twolocus_snp1+" not found\n");
  if (m2<0) error("Marker "+par::twolocus_snp2+" not found\n");
    
  vector<CSNP*>::iterator sa = SNP.begin()+m1;
  vector<CSNP*>::iterator sb = SNP.begin()+m2;
  vector<Locus*>::iterator loc1 = locus.begin()+m1;
  vector<Locus*>::iterator loc2 = locus.begin()+m2;
  vector<bool>::iterator ia1 = (*sa)->one.begin();
  vector<bool>::iterator ia2 = (*sa)->two.begin();
  vector<bool>::iterator ib1 = (*sb)->one.begin();
  vector<bool>::iterator ib2 = (*sb)->two.begin();

  vector<Individual*>::iterator person = sample.begin();
   
  int gtype = 0;

  vector_t c_all(16);
  vector_t c_case(16);
  vector_t c_control(16);
  
  while ( person != sample.end() )
    {
      
      if ( (!*ia1) && (!*ia2) ) 	
	{
	  if      ( (!*ib1) && (!*ib2) ) gtype=0;
	  else if ( (!*ib1) &&   *ib2  ) gtype=1;
	  else if (   *ib1  &&   *ib2  ) gtype=2;
	  else if (   *ib1  && (!*ib2) ) gtype=3;
	}
      else if ( (!*ia1) &&   *ia2  ) 
	{
	  if      ( (!*ib1) && (!*ib2) ) gtype=4;
	  else if ( (!*ib1) &&   *ib2  ) gtype=5;
	  else if (   *ib1  &&   *ib2  ) gtype=6;
	  else if (   *ib1  && (!*ib2) ) gtype=7;
	}
      else if (   *ia1  &&   *ia2  ) 
	{
	  if      ( (!*ib1) && (!*ib2) ) gtype=8;
	  else if ( (!*ib1) &&   *ib2  ) gtype=9;
	  else if (   *ib1  &&   *ib2  ) gtype=10;
	  else if (   *ib1  && (!*ib2) ) gtype=11;
	}
      else if (   *ia1  && (!*ia2) ) 
	{
	  if      ( (!*ib1) && (!*ib2) ) gtype=12;
	  else if ( (!*ib1) &&   *ib2  ) gtype=13;
	  else if (   *ib1  &&   *ib2  ) gtype=14;
	  else if (   *ib1  && (!*ib2) ) gtype=15;
	}

      c_all[gtype]++;
      if ( ! (*person)->missing )
	{
	  if ( (*person)->aff )
	    c_case[gtype]++;
	  else
	    c_control[gtype]++;
	}
      
      ia1++;
      ia2++;
      ib1++;
      ib2++;
      person++;
      
    }


  TWOL << "\nAll individuals\n===============\n";
  make2LTable(TWOL,*this,m1,m2,c_all,false);

  TWOL.setf(ios::fixed);
  make2LTable(TWOL,*this,m1,m2,c_all,true);
  TWOL.unsetf(ios::fixed);

  TWOL << "\nCases\n=====\n";
  make2LTable(TWOL,*this,m1,m2,c_case,false);
  TWOL.setf(ios::fixed);
  make2LTable(TWOL,*this,m1,m2,c_case,true);
  TWOL.unsetf(ios::fixed);
  TWOL << "\nControls\n========\n";
  make2LTable(TWOL,*this,m1,m2,c_control,false);
  TWOL.setf(ios::fixed);
  make2LTable(TWOL,*this,m1,m2,c_control,true);
  TWOL.unsetf(ios::fixed);

  TWOL << "\n";

  TWOL.close();

}


void Plink::extractExcludeSet(bool exclude)
{
  
  // Make map of locus name with 'l' number
  map<string,int> mlocus;
  vector<bool> del(nl_all);
  for (int l=0;l<nl_all;l++)
    {
      mlocus.insert(make_pair(locus[l]->name,l));
      if (exclude) del[l] = false;  // start off all included
      else del[l] = true;           // start off all excluded
    }
    
  map<string,int>::iterator ilocus;
  
  
  //////////////////////////////////
  // Either extract a certain "GENE"

  if (par::dump_gene)
    {
      
      readSet();
      Set S(snpset);
      
      bool found_gene = false;
      int c=0;

      for (int s=0; s<setname.size(); s++)
	{

	  if ( setname[s] == par::dump_genename )
	    {
	      // Retain just these SNPs
	      for (int l=0; l<snpset[s].size(); l++)
		del[snpset[s][l]] = false;
	      found_gene = true;
	      c = snpset[s].size();
	      break;
	    }
	  if (found_gene) break;
	}
      
      if (found_gene)
	printLOG(int2str(c)+" SNPs selected from [ " + par::dump_genename + " ]\n");
      else
	printLOG("WARNING: ** Could not find [ " + par::dump_genename + " ] in [ " + par::setfile + " ]\n");

    }

  /////////////////////////////////////
  // OR get a list of SNPs from a file

  else
    {
      string filename = par::extract_file;
      
      if (exclude)
	filename = par::exclude_file;
      
      checkFileExists(filename);  
      
      ifstream INFILE(filename.c_str(),ios::in);
      INFILE.clear();

      printLOG("Reading list of SNPs ");
      if (exclude) printLOG("to exclude [ "+par::exclude_file + " ] ... ");
      else printLOG("to extract [ "+par::extract_file + " ] ... ");
      
      int c=0;
      while (!INFILE.eof())
	{
	  string m;
	  INFILE >> m;
	  if (m=="") continue;
	  
	  ilocus = mlocus.find(m);
	  if (ilocus != mlocus.end())
	    {
	      if (exclude) del[ilocus->second] = true; 
	      else del[ilocus->second] = false; 
	      c++;
	    }
	}
      INFILE.close();
      printLOG(int2str(c)+" read\n");
    }


  ////////////////////////////////////////
  // Remove selected loci from locus list, 
  
  deleteSNPs(del);

  // Re-Read sets, as the SNP numbers will now have changed
  if (par::dump_gene)
    readSet();

}

void Plink::removeIndividuals(bool keep)
{
  

  //////////////////////////////////
  // Make map of individuals FID/IID

  map<string,int> mperson;
  map<string,int>::iterator iperson;
  vector<bool> del(n);
  for (int i=0;i<n;i++)
    {
      mperson.insert(make_pair(sample[i]->fid+"_"+sample[i]->iid,i));
      if (!keep) del[i] = false;    // start off all included
      else del[i] = true;           // start off all excluded
    }
  

  ///////////////////////////
  // Read list of individuals

  string filename = par::remove_indiv_list;
  if (keep)
    filename = par::keep_indiv_list;

  checkFileExists(filename);  
  ifstream INFILE(filename.c_str(),ios::in);
  INFILE.clear();
   
  printLOG("Reading individuals ");
  if (keep) printLOG("to keep [ "+par::keep_indiv_list + " ] ... ");
  else printLOG( "to remove [ "+par::remove_indiv_list + " ] ... ");

  int c=0;
  while (!INFILE.eof())
    {
      string fid,iid;
      INFILE >> fid >> iid;
      if (fid=="" || iid=="") continue;
      
      iperson = mperson.find(fid+"_"+iid);
      if (iperson != mperson.end())
	{
	  if (!keep) del[iperson->second] = true; 
	  else del[iperson->second] = false; 
	  c++;
	}
    }
  INFILE.close();
  printLOG(int2str(c)+" read\n");
  

  ////////////////////////////////////
  // Remove individuals as appropriate

  int n_removed = deleteIndividuals(del);

  printLOG(int2str(n_removed)+" individuals removed with ");
  if (keep) printLOG("--keep option\n");
  else printLOG("--remove option\n");
  
}



void Plink::keep2SetsForGenome()
{
  
  //////////////////////////////////
  // Make map of individuals FID/IID
  
  map<string,int> mperson;
  map<string,int>::iterator iperson;
  vector<bool> del(n);
  for (int i=0;i<n;i++)
    {
      mperson.insert(make_pair(sample[i]->fid+"_"+sample[i]->iid,i));
      del[i] = true;           // start off all excluded
    }
  

  ///////////////////////////
  // Read list of individuals

  checkFileExists(par::genome_setlist1);
  checkFileExists(par::genome_setlist2);

  ifstream INFILE1(par::genome_setlist1.c_str(),ios::in);
  INFILE1.clear();
   
  ifstream INFILE2(par::genome_setlist2.c_str(),ios::in);
  INFILE2.clear();
  
  int c=0;
  while (!INFILE1.eof())
    {
      string fid,iid;
      INFILE1 >> fid >> iid;
      if (fid=="" || iid=="") continue;
      
      iperson = mperson.find(fid+"_"+iid);
      if (iperson != mperson.end())
	{
	  del[iperson->second] = false; 
	  gset1.insert(sample[iperson->second]);
	  c++;
	}
    }
  INFILE1.close();
  printLOG(int2str(c)+" read from [ "+par::genome_setlist1 +" ]\n");
  
  c=0;
  while (!INFILE2.eof())
    {
      string fid,iid;
      INFILE2 >> fid >> iid;
      if (fid=="" || iid=="") continue;
      
      iperson = mperson.find(fid+"_"+iid);
      if (iperson != mperson.end())
	{
	  del[iperson->second] = false; 
	  gset2.insert(sample[iperson->second]);
	  c++;
	}
    }
  INFILE2.close();
  printLOG(int2str(c)+" read from [ "+par::genome_setlist2 +" ]\n");
    
  
  ////////////////////////////////////
  // Remove individuals as appropriate

  int n_removed = deleteIndividuals(del);

}



void Plink::zeroOnCluster()
{
  
  // Make map of locus name with 'l' number
  map<string,int> mlocus;
  vector<bool> del(nl_all);
  for (int l=0;l<nl_all;l++)
    mlocus.insert(make_pair(locus[l]->name,l));
  
  map<string,int>::iterator ilocus;
  
  
  ///////////////////////////////////////////
  // Get a list of SNPs/clusters from a file

  
  string filename = par::zero_cluster_filename;
  checkFileExists(filename);  
  
  ifstream INFILE(filename.c_str(),ios::in);
  INFILE.clear();

  printLOG("Reading list of SNP/clusters to zero out [ ");
  printLOG(par::zero_cluster_filename + " ]\n");
  
  
  int c=0;
  while (!INFILE.eof())
    {
      string m;
      string k;
      INFILE >> m >> k;
      
      if (m=="") continue;
      
      ilocus = mlocus.find(m);
      if (ilocus != mlocus.end())
	{
	  int l = ilocus->second;
	  int kn;
	  map<string,int>::iterator ki = kmap.find(k);

	  if ( ki == kmap.end() )
	    {
	      continue; // cluster does not exist
	      //printLOG("Cluster [ "+k+" ] not found \n");
	    }
	  else
	    {
	      int k2 = ki->second;
	      // Zero out the following
	      for (int i=0; i<n; i++)
		{
		  if ( sample[i]->sol == k2 )
		    {
		      if ( par::SNP_major ) 
			{
			  SNP[l]->one[i] = true;
			  SNP[l]->two[i] = false;
			}
		      else
			{
			  sample[i]->one[l] = true;
			  sample[i]->two[l] = false;
			}
		      c++;
		    }
		}
	    }
	}
    }
  INFILE.close();
  printLOG(int2str(c)+" genotypes zeroed\n");
}



void Plink::setObligMissing()
{

  // Temporaily use the cluster (sol) variables to store information
  // about oblig missing (i.e. call this *before* we read any 
  // subsequent cluster variables
  

  printLOG("Reading list of SNP/clusters that are obligatory missing [ ");
  printLOG(par::oblig_missing_filename + " ]\n");

  printLOG("Reading the clusters that define obligatory missingness [ ");
  printLOG(par::oblig_clusters_filename + " ]\n");

  string stored_name = par::include_cluster_filename;
  par::include_cluster_filename = par::oblig_clusters_filename;
  readClusterFile();
  par::include_cluster_filename = stored_name;

  // Make map of locus name with 'l' number

  map<string,int> mlocus;
  vector<bool> del(nl_all);
  for (int l=0;l<nl_all;l++)
    mlocus.insert(make_pair(locus[l]->name,l));
  
  map<string,int>::iterator ilocus;
    

  ///////////////////////////////////////////
  // Get a list of SNPs/clusters from a file
  
  string filename = par::oblig_missing_filename;
  checkFileExists(filename);  
  
  ifstream INFILE(filename.c_str(),ios::in);
  INFILE.clear();

  
  int c=0;
  while (!INFILE.eof())
    {
      string m;
      string k;
      INFILE >> m >> k;
      
      if (m=="") continue;
      
      ilocus = mlocus.find(m);
      if (ilocus != mlocus.end())
	{
	  int l = ilocus->second;
	  map<string,int>::iterator ki = kmap.find(k);

	  if ( ki == kmap.end() )
	    continue; // cluster does not exist	      
	  else
	    {
	      int2 p;
	      p.p1 = l;
	      p.p2 = ki->second;
	      oblig_missing.insert(p);
	      ++c;
	    }
	}
    }
  INFILE.close();
  printLOG(int2str(c)+" SNP/cluster combinations set as obligatory missing\n");
}






void Plink::flipStrand()
{

  ////////////////////////////
  // Look-up table by SNP name

  map<string,Locus*> mlocus;
  map<string,Locus*>::iterator ilocus;

  vector<Locus*>::iterator loc = locus.begin();
  while ( loc != locus.end() )
    {
      mlocus.insert(make_pair( (*loc)->name, *loc) );
      loc++;
    }  

  ///////////////////////
  // Read in SNPs to flip

  checkFileExists(par::flip_file);  
  printLOG("Reading SNPs to flip strand from [ " + par::flip_file + " ] \n");
  ifstream INFILE(par::flip_file.c_str(),ios::in);
  INFILE.clear();
  int counter = 0;

  while (!INFILE.eof())
    {
      string m;
      INFILE >> m;
      if (m=="") continue;
      
      ilocus = mlocus.find(m);
      if (ilocus != mlocus.end())
	{
	  counter++;
	  // Flip strand
	  
	  Locus * loc = ilocus->second;
	  
	  if ( loc->allele1 == "A" ) loc->allele1 = "T";
	  else if ( loc->allele1 == "C" ) loc->allele1 = "G";
	  else if ( loc->allele1 == "G" ) loc->allele1 = "C";
	  else if ( loc->allele1 == "T" ) loc->allele1 = "A";
	  else if ( loc->allele1 == "1" ) loc->allele1 = "4";
	  else if ( loc->allele1 == "2" ) loc->allele1 = "3";
	  else if ( loc->allele1 == "3" ) loc->allele1 = "2";
	  else if ( loc->allele1 == "4" ) loc->allele1 = "1";
	  
	  if ( loc->allele2 == "A" ) loc->allele2 = "T";
	  else if ( loc->allele2 == "C" ) loc->allele2 = "G";
	  else if ( loc->allele2 == "G" ) loc->allele2 = "C";
	  else if ( loc->allele2 == "T" ) loc->allele2 = "A";
	  else if ( loc->allele2 == "1" ) loc->allele2 = "4";
	  else if ( loc->allele2 == "2" ) loc->allele2 = "3";
	  else if ( loc->allele2 == "3" ) loc->allele2 = "2";
	  else if ( loc->allele2 == "4" ) loc->allele2 = "1";
	  
	}

    }

  INFILE.close();
  
  printLOG("Flipped strand of " + int2str(counter) + " SNPs\n");
  
}


void Plink::display_recoded_PEDFILE()
{
  
  string f = par::output_file_name + ".recode.ped";
  
  printLOG("Writing recoded ped file to [ " + f + " ] \n");
  ofstream PED(f.c_str(), ios::out);
  PED.clear();

  for (int i=0;i<n;i++)
    {
      Individual * person = sample[i];
      PED << person->fid << par::recode_delimit
	  << person->iid << par::recode_delimit
	  << person->pat << par::recode_delimit
	  << person->mat << par::recode_delimit
	  << person->sexcode << par::recode_delimit;

//       if ( person->missing )
// 	PED << par::missing_phenotype;
      if (par::bt)
	PED << (int)person->phenotype;
      else
	PED << person->phenotype;
      
      for (int l=0;l<nl_all;l++)
	PED << genotypeToFile( *this , i, l );

      PED << "\n";
    }
  
  PED.close();
  
  // And we also need a new map file
  
  if (par::recode_HV)
    {
      f = par::output_file_name + ".recode.info";
      printLOG("Writing HaploView-format map file [ " + f + " ] \n");
    }
  else
    {
      f = par::output_file_name + ".recode.map";
      printLOG( "Writing new map file to [ " + f + " ] \n");
    }

  ofstream MAP(f.c_str(), ios::out);
  MAP.clear();
  
  bool HV_okay = true;

  int HV_chr = locus[0]->chr;

  if ( par::recode_HV || par::recode_whap )
    {
      if ( nl_all > 5000 ) 
	printLOG(" *** WARNING :  you are exporting a large number of SNPs (>5000) for\n"
		 "                input into Haploview/WHAP -- make sure you have adequate\n"
		 "                system resources to handle this file\n\n");
    }
  
  if ( par::recode_HV)
    for (int l=0;l<nl_all;l++)
      {
	if ( locus[l]->chr != HV_chr ) 
	  HV_okay = false;
	MAP << locus[l]->name << "\t"
	    << locus[l]->bp << "\n";
      }  
  else if ( par::recode_whap )
    {
      for (int l=0;l<nl_all;l++)
	{
	  MAP << locus[l]->chr << "\t"
	      << locus[l]->name << "\t"
	      << locus[l]->bp << "\n";
	}
    }
  else
    for (int l=0;l<nl_all;l++)
      {
	MAP << locus[l]->chr << "\t"
	    << locus[l]->name << "\t"
	    << locus[l]->pos << "\t"
	    << locus[l]->bp << "\n";
      }
  

  MAP.close();

  
  if ( par::recode_HV )
    {
      if ( ! HV_okay )
	printLOG(" *** WARNING :  you've created a Haploview file containing\n"
		 " ***            SNPs from more than 1 chromosome -- these will\n"
		 " ***            not be read properly in Haploview\n\n");
    }


  if ( par::recode_whap ) 
    {
      // WHAP also needs a DAT file
      
      string f = par::output_file_name + ".recode.dat";
      printLOG("Writing whap-format DAT file [ " + f + " ] \n");
      ofstream DAT(f.c_str(), ios::out);
      DAT.clear();

      if ( par::qt ) 
	DAT << "T trait\n";
      else if ( par::coding01 )
	DAT << "B trait\n";
      else
	DAT << "A trait\n";
      
      for (int l=0;l<nl_all;l++)
	DAT << "M " << locus[l]->name << "\n";
      
      DAT.close();
            
    }

  
}


void Plink::output_fastphase_format()
{

  string f = par::output_file_name + ".recode.phase.inp";
  
  printLOG("Writing fastphase-format file to [ " + f + " ] \n");
  ofstream OUT(f.c_str(), ios::out);
  OUT.clear();

  // Sample size and number of SNPs
  
  OUT << n << "\n" << nl_all << "\nP ";

  // Positions
  
  bool output_okay = true;
  int output_chr = locus[0]->chr;
  
  for (int l=0;l<nl_all;l++)
    {
      if ( locus[l]->chr != output_chr ) 
	output_okay = false;
      OUT << locus[l]->bp << " ";
    }  
  OUT << "\n";

  if ( ! output_okay )
    printLOG(" *** WARNING :  you've created a fastphase file containing\n"
	     " ***            SNPs from more than 1 chromosome -- these will\n"
	     " ***            not be read properly in fastphase\n\n");
  

  for (int i=0;i<n;i++)
    {

      Individual * person = sample[i];

      OUT << "# ID " << person->iid << "\n";
      
      for (int l=0;l<nl_all;l++)
	{

	  bool a1 = par::SNP_major ? SNP[l]->one[i] : person->one[l];
	  bool a2 = par::SNP_major ? SNP[l]->two[i] : person->two[l];

	  if (  ! a1 )  
	    OUT << "0";
	  else 
	    {
	      if ( a2 ) 
		OUT << "1";
	      else
		OUT << "?";
	    }
	}

      OUT << "\n";
      
      // Second allele

      for (int l=0;l<nl_all;l++)
	{
	  
	  bool a1 = par::SNP_major ? SNP[l]->one[i] : person->one[l];
	  bool a2 = par::SNP_major ? SNP[l]->two[i] : person->two[l];

	  if (  ! a1  )  
	    {
	      if ( a2 ) 
		OUT << "1";
	      else 
		OUT << "0";
	    }
	  else 
	    {
	      if ( a2 ) 
		OUT << "1";
	      else
		OUT << "?";
	    }

	}
    
      // Next individual

      OUT << "\n";

    }
    
  OUT.close();

}


void Plink::output_bimbam_format()
{ 
  // One file of SNP locations (SNP BP) *.pos.txt
  // One file of phenotypes (column 6 from FAM)
  // One file containing individual IDs and genotypes (transposed format)
  //   N_ind N_snp
  //   "IND,"IID
  //   'SNP Name' genotypes... 


  // Position file

  string f = par::output_file_name + ".recode.pos.txt";
  printLOG("Writing BIMBAM position file to [ " + f + " ] \n");
  ofstream POS(f.c_str(), ios::out);
  for (int l=0;l<nl_all;l++)
    POS << locus[l]->name << " " << locus[l]->bp << "\n";
  POS.close();


  // Phenotype information

  f = par::output_file_name + ".recode.pheno.txt";
  printLOG("Writing BIMBAM phenotype file to [ " + f + " ] \n");
  ofstream PHE(f.c_str(), ios::out);

  if (par::bt)
    for (int i=0;i<n;i++)
      PHE << (int)sample[i]->phenotype << "\n";
  else
    for (int i=0;i<n;i++)
      PHE << sample[i]->phenotype << "\n";
  
  PHE.close();
  

  // Genotype file

  f = par::output_file_name + ".recode.geno.txt";
  printLOG("Writing BIMBAM genotype file to [ " + f + " ] \n");
  ofstream GEN(f.c_str(), ios::out);

  par::missing_genotype = "?";
  par::recode_delimit = ",";
  par::recode_indelimit = "";

  GEN << n << "\n"
      << nl_all << "\n"
      << "IND";
  
  for (int i=0;i<n;i++)
    GEN << "," << sample[i]->iid;
  
  GEN << "\n";

  for (int l=0;l<nl_all;l++)
    {
      GEN << locus[l]->name;
      
      for (int i=0;i<n;i++)
	GEN << genotypeToFile( *this , i, l);

      GEN << "\n";
    }
  
  GEN.close();

}

void Plink::output_structure_format()
{

  // SNP names; inter SNP distances (with -1 starting each new
  // chromosome); 


  string f = par::output_file_name + ".recode.strct_in";
  printLOG("Writing STRUCTURE format file to [ " + f + " ] \n");
  ofstream GEN(f.c_str(), ios::out);

  for (int l=0;l<nl_all;l++)
    GEN << locus[l]->name << " ";
  GEN << "\n";

  for (int l=0;l<nl_all;l++)
    GEN << 
      ( l==0 || locus[l-1]->chr != locus[l]->chr ? 
	-1 :
	locus[l]->bp - locus[l-1]->bp ) 
	<< " ";
  GEN << "\n";

  par::missing_genotype = "0";
  par::recode_delimit = " ";
  par::recode_indelimit = " ";
  par::recode_12 = true;

  map<string,int> fmap;
  int cnt = 0;
  for (int i=0;i<n;i++)
    {
      Individual * person = sample[i];
      if ( fmap.find( person->fid ) == fmap.end() )
	fmap.insert(make_pair(person->fid,++cnt));
    }

  for (int i=0;i<n;i++)
    {
      GEN << sample[i]->iid << " "
	  << fmap.find(sample[i]->fid)->second;
      
      for (int l=0;l<nl_all;l++)
	GEN << genotypeToFile( *this , i, l);
      
      GEN << "\n";
    }
  
  GEN.close();


//   GEN << n << "\n"
//       << nl_all << "\n"
//       << "IND";
  
}


void Plink::display_recoded_PEDFILE_transpose()
{

  string f = par::output_file_name + ".recode.tped";
  printLOG("Writing transposed ped file to [ " + f + " ] \n");
  ofstream PED(f.c_str(), ios::out);
  PED.clear();

  for (int l=0;l<nl_all;l++)
    {
      PED << locus[l]->chr << par::recode_delimit
 	  << locus[l]->name << par::recode_delimit
 	  << locus[l]->pos << par::recode_delimit
	  << locus[l]->bp;
      
      for (int i=0;i<n;i++)
	PED << genotypeToFile( *this , i, l);

      PED << "\n";
    }
  
  PED.close();
  

  // And we also need a new family info file

  f = par::output_file_name + ".recode.tfam";

  printLOG( "Writing family information to [ " + f + " ] \n");  
  ofstream FAM(f.c_str(), ios::out);
  FAM.clear();
  
  for (int i=0;i<n;i++)
    {
      Individual * person = sample[i];
      FAM << person->fid << par::recode_delimit
	  << person->iid << par::recode_delimit
	  << person->pat << par::recode_delimit
	  << person->mat << par::recode_delimit
	  << person->sexcode << par::recode_delimit;
      
//       if ( person->missing )
// 	PED << par::missing_phenotype;
      if (par::bt)
	FAM << (int)person->phenotype;
      else
	FAM << person->phenotype;
      
      FAM << "\n";
    }
  
  FAM.close();
  
}

void Plink::display_recoded_PEDFILE_AD()
{

  string f = par::output_file_name + ".recode.raw";
  printLOG( "Writing recoded file to [ " + f + " ] \n");
  ofstream PED(f.c_str(), ios::out);
  PED.clear();
  
  // Header row
  PED << "FID" 
      << par::recode_delimit << "IID"
      << par::recode_delimit << "PAT"
      << par::recode_delimit << "MAT"
      << par::recode_delimit << "SEX"
      << par::recode_delimit << "PHENOTYPE";

  for (int l=0;l<nl_all;l++)
    {
      PED << par::recode_delimit << locus[l]->name+"_A";
      if ( ! par::recode_AD_Aonly )
	PED << par::recode_delimit << locus[l]->name+"_D";
    }
  
  PED << "\n";
  
  for (int i=0;i<n;i++)
    {
      Individual * person = sample[i];
      PED << person->fid 
	  << par::recode_delimit << person->iid 
	  << par::recode_delimit << person->pat 
	  << par::recode_delimit << person->mat 
	  << par::recode_delimit << person->sexcode;

//       if ( person->missing )
// 	PED << par::recode_delimit << par::missing_phenotype;
      if (par::bt)
	PED << par::recode_delimit << (int)person->phenotype;
      else
	PED << par::recode_delimit << person->phenotype;

      string g0 = par::recode_AD_Aonly ? 
	par::recode_delimit + "0" 
	: par::recode_delimit + "0" + par::recode_delimit + "0";
	  
      string g1 = par::recode_AD_Aonly ? 
	par::recode_delimit + "1" 
	: par::recode_delimit + "1" + par::recode_delimit + "1";
	  
      string g2 = par::recode_AD_Aonly ? 
	par::recode_delimit + "2" 
	: par::recode_delimit + "2" + par::recode_delimit + "0";
	  
      string gX = par::recode_AD_Aonly ? 
	par::recode_delimit + "NA" 
	: par::recode_delimit + "NA" + par::recode_delimit + "NA";
      
      
      for (int l=0;l<nl_all;l++)
	{
	  
	  bool a1 = par::SNP_major ? 
	    SNP[l]->one[i] : 
	    sample[i]->one[l];
	  
	  bool a2 = par::SNP_major ? 
	    SNP[l]->two[i] : 
	    sample[i]->two[l];
	  
	  if (par::recode_AD_fixed && locus[l]->allele1=="1")
	    {
	      if ( (!a1) && (!a2) )  
		PED << g2;
	      else if ( (!a1) && a2) 
		PED << g1;
	      else if ( a1 && a2 ) 
		PED << g0;
	      else 
		PED << gX;
	    }
	  else
	    {
	      if ( (!a1) && (!a2) )  
		PED << g0;
	      else if ( (!a1) && a2) 
		PED << g1;
	      else if ( a1 && a2 ) 
		PED << g2;
	      else 
		PED << gX;	      
	    }
	}

      
      PED << "\n";
    }
  
  PED.close();

}



void Plink::write_covariates()
{
  
  printLOG( "Writing covariate information to [ " 
	    + par::output_file_name + ".cov ] \n");
  ofstream COV((par::output_file_name+".cov").c_str(), ios::out);
  
  // Always write a header row
  COV << "FID IID ";
  if (par::dump_covar_with_phenotype)
    COV << "PAT MAT SEX PHENOTYPE ";  
  for (int c=0; c<par::clist_number; c++)
    COV << clistname[c] << " ";
  COV << "\n";
  
  // For each individual
  for (int i=0;i<n;i++)
    {
      Individual * person = sample[i];
      
      COV << person->fid << " "
	  << person->iid << " ";

      if (par::dump_covar_with_phenotype)
	{
	  COV << person->pat << " "
	      << person->mat << " "
	      << person->sex << " ";
	  
	  if ( person->missing )
	    COV << par::missing_phenotype << " ";
	  else
	    COV << person->phenotype << " ";
	}

      for (int c=0; c<par::clist_number; c++)
	COV << person->clist[c] << " ";
      
      COV << "\n";
    }
  COV.close();
}

void Plink::write_BITFILE()
{

  printLOG( "Writing pedigree information to [ " 
	    + par::output_file_name + ".fam ] \n");
  ofstream BIT((par::output_file_name+".fam").c_str(), ios::out);
  // For each individual

  for (int i=0;i<n;i++)
    {
      Individual * person = sample[i];
      BIT << person->fid << " "
	  << person->iid << " "
	  << person->pat << " "
	  << person->mat << " "
	  << person->sexcode << " ";

//       if ( person->missing )
// 	BIT << par::missing_phenotype << "\n";
      if (par::bt)
	BIT << (int)person->phenotype << "\n";
      else
	BIT << person->phenotype << "\n";
    }
  BIT.clear();
  BIT.close();

  printLOG( "Writing map (extended format) information to [ " 
	    + par::output_file_name + ".bim ] \n");
  BIT.open((par::output_file_name+".bim").c_str(), ios::out);
  
  for (int l=0;l<nl_all;l++)
    {

      if (locus[l]->allele1=="")
 	{
 	  if (locus[l]->allele2!="0") locus[l]->allele1="0";
 	  else locus[l]->allele1="X";
 	}
      
      if (locus[l]->allele2=="")
	locus[l]->allele2="0";
      
      BIT << locus[l]->chr << "\t"
	  << locus[l]->name << "\t"
	  << locus[l]->pos << "\t"
	  << locus[l]->bp << "\t" 
	  << locus[l]->allele1 << "\t"
	  << locus[l]->allele2 << "\n";
    }
  
  BIT.clear();
  BIT.close();


  //////////////////////////////////////
  // Save genotype data in BITFILE format

  printLOG("Writing genotype bitfile to [ " 
	   + par::output_file_name + ".bed ] \n");
    
  BIT.open((par::output_file_name+".bed").c_str(), ios::out | ios::binary);
  
  if (par::out_SNP_major)
    printLOG("Using (default) SNP-major mode\n");
  else
    printLOG("Using individual-major mode\n");

  bitset<8> b;
  char ch[1];


  // Magic numbers for .bed file: 00110110 11011000 = v1.00 bed file

  b.reset();
  b.set(2);  b.set(3);  b.set(5);  b.set(6);
  ch[0] = (char)b.to_ulong();
  BIT.write(ch,1);

  b.reset();
  b.set(0);  b.set(1);  b.set(3);  b.set(4);
  ch[0] = (char)b.to_ulong();
  BIT.write(ch,1);


  // BIT represents status of SNP-major (true) or Ind-major (false)

  b.reset();
  if (par::out_SNP_major)
    b.set(0);
  ch[0] = (char)b.to_ulong();
  BIT.write(ch,1);
  
  
  //////////////////////////////////////////
  // Now consider genotypes: SNP-major mode
  
  if (par::out_SNP_major)
    {  
      
      // Create the SNP-major dataset, if need be
      if (!par::SNP_major) Ind2SNP();

      vector<CSNP*>::iterator s = SNP.begin();
         
      // Outer loop over SNPs
      while ( s != SNP.end() )
	{
	  
	  vector<bool>::iterator i1 = (*s)->one.begin();
	  vector<bool>::iterator i2 = (*s)->two.begin();
	  //	  vector<Individual*>::iterator person = sample.begin();

	  // Inner loop over individuals
	  while ( i1 != (*s)->one.end() )
	    {
	      bitset<8> b;
	      b.reset();
	      int c=0;      
	      
	      while (c<8 && i1 != (*s)->one.end() )
		{
		  if ( *(i1) ) b.set(c);
		  i1++;
		  c++;

		  if ( *(i2) ) b.set(c);		  
		  i2++;
		  c++;

		  //person++;
		}
	      
	      char ch[1];
	      ch[0] = (char)b.to_ulong();
	      BIT.write(ch,1);
	      
	    }

	  // next SNP
	  s++;
	}
    }
  

  ////////////////////////////////////////////////////////////
  // Alternatively, consider genotypes: Individual-major mode

  else
    {

      // Create the individual-major dataset, if need be
      if (par::SNP_major) SNP2Ind();
      
      vector<Individual*>::iterator person = sample.begin();
            
      // Outer loop over individuals
      while ( person != sample.end() )
	{
	  
	  vector<bool>::iterator i1 = (*person)->one.begin();
	  vector<bool>::iterator i2 = (*person)->two.begin();
	  
	  // Inner loop over SNPs
	  while ( i1 != (*person)->one.end() )
	    {
	      
	      bitset<8> b;
	      b.reset();
	      int c=0;      
	      
	      while (c<8 && i1 != (*person)->one.end() )
		{
		  if ( *(i1) ) b.set(c);
		  i1++;
		  c++;
		  if ( *(i2) ) b.set(c);		  
		  i2++;
		  c++;
		}
	      
	      char ch[1];
	      ch[0] = (char)b.to_ulong();
	      BIT.write(ch,1);
	      
	    }
	  
	  // next person
	  person++;
	}
      

    }


  BIT.close();
  
}



void Plink::summaryIBSsegments(Perm & perm)
{

  vector<int> coverage_conc_aff(nl_all,0);
  vector<int> coverage_conc_unaff(nl_all,0);
  vector<int> coverage_disc(nl_all,0);
    
  vector<Segment>::iterator s = segment.begin();
  
  while ( s != segment.end() )
    {    
      
      if (s->p1->aff == s->p2->aff)
	{
	  if ( s->p1->aff) 
	    for (int l = s->start ; l <= s->finish; l++) coverage_conc_aff[l]++;
	  else
	    for (int l = s->start ; l <= s->finish; l++) coverage_conc_unaff[l]++;
	}
      else
	for (int l = s->start ; l <= s->finish; l++) coverage_disc[l]++;
              
      s++;
    }
    
  
  string f = par::output_file_name;

  if ( par::segment_output ) 
    f += ".segment.summary";
  else if ( par::homo_run )
    f += ".hom.summary";
  
  else if ( par::ibs_2only )
    f += ".ibs2.summary";
  else
    f += ".ibs.summary";
  
  ofstream SIBS;
  SIBS.open( f.c_str() , ios::out );
  

  //////////////////////////////
  // And display

  if ( par::homo_run )
    SIBS << setw(4) << "CHR" << " " 
	 << setw(par::pp_maxsnp) << "SNP" << " " 
	 << setw(8) << "UNAFF" << " " 
	 << setw(8) << "AFF" << "\n"; 
  else
    SIBS << setw(4) << "CHR" << " " 
	 << setw(par::pp_maxsnp) << "SNP" << " " 
	 << setw(8) << "CONU" << " " 
	 << setw(8) << "DISC" << " " 
	 << setw(8) << "CONA" << "\n"; 
  
  for (int l=0; l<nl_all; l++)
    {
      
  if ( par::homo_run )
    SIBS << setw(4) << locus[l]->chr << " " 
	 << setw(par::pp_maxsnp) << locus[l]->name << " " 
	 << setw(8) << coverage_conc_unaff[l] << " " 
	 << setw(8) << coverage_conc_aff[l] << "\n"; 
  else
    SIBS << setw(4) << locus[l]->chr << " " 
	 << setw(par::pp_maxsnp) << locus[l]->name << " " 
	 << setw(8) << coverage_conc_unaff[l] << " " 
	 << setw(8) << coverage_disc[l] << " " 
	 << setw(8) << coverage_conc_aff[l] << "\n"; 
      
    }

  SIBS.close();


  //////////////////////////////////
  // Permutation test, if C/C given?

  if (!par::permute) return;
  
  if (par::homo_run)
    homozygousSegmentPermutationTest(perm,f,coverage_conc_aff,coverage_conc_unaff);
  else
    segmentPermutationTest(perm,false,f,coverage_conc_aff,coverage_disc,coverage_conc_unaff);

  return;

}


void Plink::summaryIBDsegments(Perm & perm)
{

  // Number of map positions
  // (extra position for global IBD)
  int npos = m1.size() - 1;
  
  vector<int> coverage_conc_aff(nl,0);
  vector<int> coverage_conc_unaff(nl,0);
  vector<int> coverage_disc(nl,0);
  
  // pihat[pair][position]
  // check if, e.g. pihat>0.2 (i.e. atleast 80% chance of IBD1 sharing)
  // level set in options.cpp

      
  for (int l=0;l<npos;l++)
    {
      
      double p1, p2;
    
      if (m1[l]==-1) 
	{
	  p1 = locus[par::run_start]->pos - par::fringe;
	}
      else 
	{
	  p1 = locus[m1[l]]->pos;
	}
      
      if (m2[l]==-1) 
	{
	  p2 = locus[par::run_end]->pos + par::fringe;
	}
      else 
	{
	  p2 = locus[m2[l]]->pos;
	}
      
      if (m1[l]==-1 && m2[l]==-1)
	{
	  p1 = p2 = 0;
	}
      
      double d1 = p1 + pos[l] * (p2-p1);

      
      vector<Segment>::iterator s = segment.begin();
      
      while ( s != segment.end() )
	{    
	  
	  if ( locus[s->start]->chr != par::run_chr )
	    {
	      s++;
	      continue;
	    }
	  
	  
	  // In segment?
	  if (locus[s->start]->pos < d1 && locus[s->finish]->pos > d1  )
	    {
	      
	      if (s->p1->aff == s->p2->aff)
		{
		  if ( s->p1->aff) 
		    coverage_conc_aff[l]++;
		  else
		    coverage_conc_unaff[l]++;
		}
	      else
		coverage_disc[l]++;
	      
	    }
	  
	  s++;
	}
      
    }
  
  
  string f = par::output_file_name + ".segment.summary";

  ofstream SSEG;
  if (par::segment_output_started)
     SSEG.open( f.c_str() , ios::app );
  else
     SSEG.open( f.c_str() , ios::out );
  par::segment_output_started = true;

  //////////////////////////////
  // And display
  
  for (int l=0; l<npos; l++)
    {
      
      double p1, p2;
      string n1, n2;
      double fr=0;
      
      ////////////////////////////////
      // Single and multipoint output
      
      if (m1[l]==-1) 
	{
	  p1 = locus[par::run_start]->pos - par::fringe;
	  n1 = "fringe";
	}
      else 
	{
	  p1 = locus[m1[l]]->pos;
	  n1 = locus[m1[l]]->name;
	  fr = locus[m1[l]]->freq;
	}
      
      if (m2[l]==-1)
	{
	  p2 = locus[par::run_end]->pos + par::fringe;
	  n2 = "fringe";
	}
      else 
	{
	  p2 = locus[m2[l]]->pos;
	  n2 = locus[m2[l]]->name;
	}
      
      if (m1[l]==-1 && m2[l]==-1)
	{
	  p1 = p2 = 0;
	  n1 = "Genomewide";
	  n2 = "IBD";
	}

      double d1 = p1 + pos[l] * (p2-p1);

      SSEG << par::run_chr << " " 
	   << n1 << " " 
	   << n2 << " " 
	   << d1 << "  "
	   << fr << " " 
	   << coverage_conc_unaff[l] << " " 
	   << coverage_disc[l] << " " 
	   << coverage_conc_aff[l] << "\n"; 

    }

  SSEG.close();

  
  ////////////////////////
  // C/C permutation test?

  if (!par::permute) return;

  segmentPermutationTest(perm,true,f,coverage_conc_aff,coverage_disc,coverage_conc_unaff);

}



void Plink::summaryIBD()
{

  int npos = pihat[0].size();
  int npair = pihat.size();
  
  vector<int> coverage(npos,0);
  
  vector<int> coverage_conc_aff(npos,0);
  vector<int> coverage_conc_unaff(npos,0);
  vector<int> coverage_disc(npos,0);
  
  // pihat[pair][position]
  // check if, e.g. pihat>0.2 (i.e. atleast 80% chance of IBD1 sharing)
  // level set in options.cpp

  for (int i=0;i<pihat.size();i++)
    for (int j=0;j<npos;j++)
      {
	if (pihat[i][j] > par::IBD_threshold )
	  {
	    coverage[j]++;
	    
	    if (sample[pair1[i]]->phenotype 
		!= sample[pair2[i]]->phenotype) 
	      coverage_disc[j]++;
	    
	    if (sample[pair1[i]]->phenotype==1 
		&& sample[pair2[i]]->phenotype==1) 
	      coverage_conc_unaff[j]++;

	    if (sample[pair1[i]]->phenotype==2 
		&& sample[pair2[i]]->phenotype==2) 
	      coverage_conc_aff[j]++;
	  }
      }


  //////////////////////////////
  // And display
  
  for (int l=0; l<npos; l++)
    {
      
      double p1, p2;
      string n1, n2;
      double fr = 0;

      ////////////////////////////////
      // Single and multipoint output
      
      if (m1[l]==-1) 
	{
	  p1 = locus[par::run_start]->pos - par::fringe;
	  n1 = "fringe";
	}
      else 
	{
	  p1 = locus[m1[l]]->pos;
	  n1 = locus[m1[l]]->name;
	  fr = locus[m1[l]]->freq;
	}
      
      if (m2[l]==-1) 
	{
	  p2 = locus[par::run_end]->pos + par::fringe;
	  n2 = "fringe";
	}
      else 
	{
	  p2 = locus[m2[l]]->pos;
	  n2 = locus[m2[l]]->name;
	}
      
      if (m1[l]==-1 && m2[l]==-1)
	{
	  p1 = p2 = 0;
	  n1 = "Genomewide";
	  n2 = "IBD";
	}
      
      double d1 = p1 + pos[l] * (p2-p1);
      
      cout << "C " 
	   << par::run_chr << " " 
	   << n1 << " " 
	   << n2 << " " 
	   << d1 << "  "
	   << fr << " " 
	   << coverage[l] << " " 
	   << coverage_conc_unaff[l] << " " 
	   << coverage_disc[l] << " " 
	   << coverage_conc_aff[l] << " " 
	   << coverage[l]/(double)npair << "\n"; 
    }
}


void Plink::displayGMULTI(Individual * p1, Individual * p2, int l, ofstream & GMULTI)
{

  Locus * loc = locus[l];

  bool a1 = p1->one[l];
  bool a2 = p1->two[l];

  bool b1 = p2->one[l];
  bool b2 = p2->two[l];

  bool miss = false;
  
  GMULTI << setw(par::pp_maxfid) << p1->fid << " "
	 << setw(par::pp_maxiid) << p1->iid << " "
	 << setw(par::pp_maxfid) << p2->fid << " "
	 << setw(par::pp_maxiid) << p2->iid << " "
	 << setw(par::pp_maxsnp) << loc->name << " ";

  // Minor allele and frequency 

  GMULTI << setw(2) << loc->allele1 << " "
	 << setw(8) << loc->freq << " ";

  // Genotypes

  if ((!a1) && (!a2)) 
    GMULTI << setw(4) 
	   << loc->allele1+"/"+loc->allele1 
	   << " ";
  else if ((!a1) && a2) 
    GMULTI << setw(4) 
	   << loc->allele1+"/"+loc->allele2
	   << " ";
  else if (a1 && a2) 
    GMULTI << setw(4) 
	   << loc->allele2+"/"+loc->allele2
	   << " ";
  else 
    {
      GMULTI << setw(4) << "0/0" << " ";
      miss = true; 
    }
  
  if ((!b1) && (!b2)) 
    GMULTI << setw(4) 
	   << loc->allele1+"/"+loc->allele1;
  else if ((!b1) && b2) 
    GMULTI << setw(4) 
	   << loc->allele1+"/"+loc->allele2;
  else if (b1 && b2) 
    GMULTI << setw(4) 
	   << loc->allele2+"/"+loc->allele2;
  else 
    {
      GMULTI << setw(4) << "0/0";
      miss = true;
    }

  // IBS count

  if (miss)
    GMULTI << setw(3) << "NA" << " ";
  else
    {
      int ibs=0;
      if (a1==b1) ibs++;
      if (a2==b2) ibs++;
      GMULTI << setw(3) << ibs << " ";
    }

  GMULTI << "\n";
  
}
